import 'dart:async';
import 'dart:ui' as ui;

import 'package:flutter/services.dart';

import 'interfaces/pdf_render.dart';
import 'interfaces/pdf_render_platform_interface.dart';
import 'utils/ffi_wrapper.dart';

const MethodChannel _channel = MethodChannel('pdf_render');

class PdfRenderPlatformMethodChannel extends PdfRenderPlatform {
  PdfDocument _open(Map<dynamic, dynamic> obj, String sourceName) {
    final pageCount = obj['pageCount'] as int;
    return PdfDocumentMethodChannel._(
      sourceName: sourceName,
      docId: obj['docId'] as int,
      pageCount: pageCount,
      verMajor: obj['verMajor'] as int,
      verMinor: obj['verMinor'] as int,
      isEncrypted: obj['isEncrypted'] as bool,
      allowsCopying: obj['allowsCopying'] as bool,
      allowsPrinting: obj['allowsPrinting'] as bool,
      //isUnlocked: obj['isUnlocked'] as bool
    );
  }

  /// Opening the specified file.
  @override
  Future<PdfDocument> openFile(String filePath) async {
    return _open(await _channel.invokeMethod('file', filePath), filePath);
  }

  /// Opening the specified asset.
  @override
  Future<PdfDocument> openAsset(String name) async {
    return _open(await _channel.invokeMethod('asset', name), 'asset:$name');
  }

  /// Opening the PDF on memory.
  @override
  Future<PdfDocument> openData(Uint8List data) async {
    return _open(await _channel.invokeMethod('data', data), 'memory:');
  }

  /// Create a new Flutter [Texture]. The object should be released by calling [dispose] method after use it.
  @override
  Future<PdfPageImageTextureMethodChannel> createTexture() async {
    final texId = await _channel.invokeMethod<int>('allocTex');
    return PdfPageImageTextureMethodChannel._(texId: texId!);
  }
}

/// Handles PDF document loaded on memory.
class PdfDocumentMethodChannel extends PdfDocument {
  /// Document-ID that uniquely identifies the current instance.
  final int docId;

  final List<PdfPage?> _pages;

  PdfDocumentMethodChannel._({
    required super.sourceName,
    required this.docId,
    required super.pageCount,
    required super.verMajor,
    required super.verMinor,
    required super.isEncrypted,
    required super.allowsCopying,
    required super.allowsPrinting,
  }) : _pages = List<PdfPage?>.filled(pageCount, null);

  @override
  Future<void> dispose() async {
    await _channel.invokeMethod('close', docId);
  }

  /// Get page object. The first page is 1.
  @override
  Future<PdfPage> getPage(int pageNumber) async {
    if (pageNumber < 1 || pageNumber > pageCount) {
      throw RangeError.range(pageNumber, 1, pageCount, 'pageNumber');
    }
    var page = _pages[pageNumber - 1];
    if (page == null) {
      var obj = (await _channel.invokeMethod<Map<dynamic, dynamic>>(
          'page', {"docId": docId, "pageNumber": pageNumber}))!;
      page = _pages[pageNumber - 1] = PdfPageMethodChannel._(
        document: this,
        pageNumber: pageNumber,
        width: obj['width'] as double,
        height: obj['height'] as double,
      );
    }
    return page;
  }

  @override
  bool operator ==(dynamic other) =>
      other is PdfDocumentMethodChannel && other.docId == docId;

  @override
  int get hashCode => docId;

  @override
  String toString() => sourceName;
}

/// Handles a PDF page in [PDFDocument].
class PdfPageMethodChannel extends PdfPage {
  PdfPageMethodChannel._(
      {required PdfDocumentMethodChannel super.document,
      required super.pageNumber,
      required super.width,
      required super.height});

  @override
  Future<PdfPageImage> render({
    int x = 0,
    int y = 0,
    int? width,
    int? height,
    double? fullWidth,
    double? fullHeight,
    bool? backgroundFill,
    bool? allowAntialiasingIOS,
  }) async {
    return PdfPageImageMethodChannel._render(
      document as PdfDocumentMethodChannel,
      pageNumber,
      x: x,
      y: y,
      width: width,
      height: height,
      fullWidth: fullWidth,
      fullHeight: fullHeight,
      backgroundFill: backgroundFill,
      allowAntialiasingIOS: allowAntialiasingIOS,
    );
  }

  @override
  bool operator ==(dynamic other) =>
      other is PdfPageMethodChannel &&
      other.document == document &&
      other.pageNumber == pageNumber;

  @override
  int get hashCode => document.hashCode ^ pageNumber;

  @override
  String toString() => '$document:page=$pageNumber';

  @override
  Future<bool> renderOnTexture(
      {required PdfPageImageTexture texture,
      int x = 0,
      int y = 0,
      required int width,
      required int height,
      double? fullWidth,
      double? fullHeight,
      bool backgroundFill = true,
      bool allowAntialiasingIOS = true}) async {
    final tex = texture as PdfPageImageTextureMethodChannel;
    final result = (await _channel.invokeMethod<int>('updateTex', {
      'texId': tex.texId,
      'width': width,
      'height': height,
      'srcX': x,
      'srcY': y,
      'fullWidth': fullWidth,
      'fullHeight': fullHeight,
      'backgroundFill': backgroundFill,
      'allowAntialiasingIOS': allowAntialiasingIOS,
    }))!;
    if (result >= 0) {
      tex._texWidth = width;
      tex._texHeight = height;
    }
    return result >= 0;
  }
}

class PdfPageImageMethodChannel extends PdfPageImage {
  final Uint8List _pixels;
  Pointer<Uint8>? _buffer;
  ui.Image? _imageCached;

  PdfPageImageMethodChannel._(
      {required super.pageNumber,
      required super.x,
      required super.y,
      required super.width,
      required super.height,
      required super.fullWidth,
      required super.fullHeight,
      required super.pageWidth,
      required super.pageHeight,
      required Uint8List pixels,
      Pointer<Uint8>? buffer})
      : _pixels = pixels,
        _buffer = buffer;

  /// RGBA pixels in byte array.
  @override
  Uint8List get pixels => _pixels;

  /// Pointer to the internal RGBA image buffer if available; the size is calculated by `width*height*4`.
  @override
  Pointer<Uint8>? get buffer => _buffer;

  @override
  void dispose() {
    _imageCached?.dispose();
    _imageCached = null;
    if (_buffer != null) {
      _channel.invokeMethod('releaseBuffer', _buffer!.address);
      _buffer = null;
    }
  }

  /// Create cached [Image] for the page.
  @override
  Future<ui.Image> createImageIfNotAvailable() async {
    _imageCached ??= await _decodeRgba(width, height, _pixels);
    return _imageCached!;
  }

  /// Get [Image] for the object if available; otherwise null.
  /// If you want to ensure that the [Image] is available, call [createImageIfNotAvailable].
  @override
  ui.Image? get imageIfAvailable => _imageCached;

  @override
  Future<ui.Image> createImageDetached() async =>
      await _decodeRgba(width, height, _pixels);

  static Future<PdfPageImage> _render(
    PdfDocumentMethodChannel document,
    int pageNumber, {
    int? x,
    int? y,
    int? width,
    int? height,
    double? fullWidth,
    double? fullHeight,
    bool? backgroundFill,
    bool? allowAntialiasingIOS,
  }) async {
    final obj = (await _channel.invokeMethod<Map<dynamic, dynamic>>('render', {
      'docId': document.docId,
      'pageNumber': pageNumber,
      'x': x,
      'y': y,
      'width': width,
      'height': height,
      'fullWidth': fullWidth,
      'fullHeight': fullHeight,
      'backgroundFill': backgroundFill,
      'allowAntialiasingIOS': allowAntialiasingIOS,
    }))!;

    final retWidth = obj['width'] as int;
    final retHeight = obj['height'] as int;
    Pointer<Uint8>? ptr;
    var pixels = obj['data'] as Uint8List? ??
        () {
          final addr = obj['addr'] as int;
          final size = obj['size'] as int;
          ptr = Pointer<Uint8>.fromAddress(addr);
          return ptr!.asTypedList(size);
        }();

    return PdfPageImageMethodChannel._(
        pageNumber: obj['pageNumber'] as int,
        x: obj['x'] as int,
        y: obj['y'] as int,
        width: retWidth,
        height: retHeight,
        fullWidth: obj['fullWidth'] as double,
        fullHeight: obj['fullHeight'] as double,
        pageWidth: obj['pageWidth'] as double,
        pageHeight: obj['pageHeight'] as double,
        pixels: pixels,
        buffer: ptr);
  }

  /// Decode RGBA raw image from native code.
  static Future<ui.Image> _decodeRgba(int width, int height, Uint8List pixels) {
    final comp = Completer<ui.Image>();
    ui.decodeImageFromPixels(pixels, width, height, ui.PixelFormat.rgba8888,
        (image) => comp.complete(image));
    return comp.future;
  }
}

/// Very limited support for Flutter's [Texture] based drawing.
/// Because it does not transfer the rendered image via platform channel,
/// it could be faster and more efficient than the [PdfPageImage] based rendering process.
class PdfPageImageTextureMethodChannel extends PdfPageImageTexture {
  @override
  final int texId;

  PdfPageImageTextureMethodChannel._({required this.texId});

  /// Release the object.
  @override
  Future<void> dispose() => _channel.invokeMethod('releaseTex', texId);

  int? _texWidth;
  int? _texHeight;

  @override
  int? get texWidth => _texWidth;
  @override
  int? get texHeight => _texHeight;
  @override
  bool get hasUpdatedTexture => _texWidth != null;

  @override
  bool operator ==(Object other) {
    return other is PdfPageImageTextureMethodChannel && other.texId == texId;
  }

  @override
  int get hashCode => texId.hashCode;

  Future<void> paint(
    int width,
    int height, {
    bool backgroundFill = true,
    required void Function(
      Pointer<Uint8> argbPixels,
      int width,
      int height,
      int stride,
      int size,
    ) paint,
  }) async {
    final obj = await _channel
        .invokeMethod<Map<dynamic, dynamic>>('lockTextureBufferARGB', {
      'width': width,
      'height': height,
      'backgroundFill': backgroundFill,
    });
    if (obj == null) {
      throw Exception('Failed to lock texture buffer.');
    }

    paint(
      Pointer<Uint8>.fromAddress(obj['buffer'] as int),
      obj['width'] as int,
      obj['height'] as int,
      obj['stride'] as int,
      obj['size'] as int,
    );
    await _channel.invokeMethod('unlockTextureBufferARGB', {
      'texId': texId,
      'globalRef': obj['globalRef'],
    });
  }
}
