// ignore_for_file: public_member_api_docs, sort_constructors_first
import 'dart:async';
import 'dart:ui' as ui;

import 'package:ffi/ffi.dart';
import 'package:flutter/services.dart';
import 'package:pdf_render/src/interfaces/pdf_render.dart';
import 'package:pdf_render/src/pdf_render_method_channel.dart';
import 'package:pdf_render/src/utils/ffi_wrapper.dart';
import 'package:pdfium_bindings/pdfium_bindings.dart';

import '../interfaces/pdf_render_platform_interface.dart';

final libpdfium = DynamicLibrary.open('libpdfium.so');
final pdfium = PDFiumBindings(libpdfium);
bool _initialized = false;

void _init() {
  if (_initialized) return;
  using((arena) {
    final config =
        arena.allocate<FPDF_LIBRARY_CONFIG>(sizeOf<FPDF_LIBRARY_CONFIG>());
    config.ref.version = 2;
    config.ref.m_pUserFontPaths = Pointer.fromAddress(0);
    config.ref.m_pIsolate = Pointer.fromAddress(0);
    config.ref.m_v8EmbedderSlot = 0;
    pdfium.FPDF_InitLibraryWithConfig(config);
  });
  _initialized = true;
}

class PdfRenderPlatformPdfium extends PdfRenderPlatform {
  @override
  Future<PdfPageImageTexture> createTexture() =>
      PdfRenderPlatform.stockImpl.createTexture();

  @override
  Future<PdfDocument> openAsset(String name) async {
    final data = await rootBundle.load(name);
    return await _openData(data.buffer.asUint8List(), 'asset:$name');
  }

  @override
  Future<PdfDocument> openData(Uint8List data) =>
      _openData(data, 'memory-${data.hashCode}');

  @override
  Future<PdfDocument> openFile(String filePath) async {
    _init();
    return using((arena) {
      return PdfDocumentPdfium.fromPdfDocument(
        pdfium.FPDF_LoadDocument(
            filePath.toUtf8(arena), Pointer.fromAddress(0)),
        sourceName: filePath,
      );
    });
  }

  Future<PdfDocument> _openData(Uint8List data, String sourceName) async {
    _init();
    return using((arena) {
      final mem = data.allocateCopy();
      return PdfDocumentPdfium.fromPdfDocument(
        pdfium.FPDF_LoadMemDocument(mem, data.length, Pointer.fromAddress(0)),
        sourceName: sourceName,
        disposeCallback: () => calloc.free(mem),
      );
    });
  }
}

extension FpdfMemoryExt on Uint8List {
  Pointer<Void> allocateCopy() {
    final ptr = calloc<Uint8>(length);
    ptr.asTypedList(length).setAll(0, this);
    return ptr.cast<Void>();
  }
}

extension FpdfUtf8StringExt on String {
  Pointer<Int8> toUtf8(Arena arena) =>
      Pointer.fromAddress(toNativeUtf8(allocator: arena).address);
}

typedef FpdfPdfDocument = Pointer<fpdf_document_t__>;
typedef FpdfPdfPage = Pointer<fpdf_page_t__>;

class PdfDocumentPdfium extends PdfDocument {
  final FpdfPdfDocument doc;
  final List<FpdfPdfPage> _pages;
  final void Function()? disposeCallback;

  PdfDocumentPdfium._(
    this.doc, {
    required super.sourceName,
    required super.pageCount,
    required super.verMajor,
    required super.verMinor,
    required super.isEncrypted,
    required super.allowsCopying,
    required super.allowsPrinting,
    required List<FpdfPdfPage> pages,
    this.disposeCallback,
  }) : _pages = pages;

  static PdfDocument fromPdfDocument(FpdfPdfDocument doc,
      {required String sourceName, void Function()? disposeCallback}) {
    return using((arena) {
      final pageCount = pdfium.FPDF_GetPageCount(doc);
      final version = arena.allocate<Int32>(sizeOf<Int32>());
      pdfium.FPDF_GetFileVersion(doc, version);
      final permissions = pdfium.FPDF_GetDocPermissions(doc);
      return PdfDocumentPdfium._(
        doc,
        sourceName: sourceName,
        pageCount: pageCount,
        verMajor: version.value ~/ 10,
        verMinor: version.value % 10,
        isEncrypted: pdfium.FPDF_GetSecurityHandlerRevision(doc) != -1,
        allowsCopying: permissions & 32 != 0,
        allowsPrinting: permissions & 8 != 0,
        pages: List<FpdfPdfPage>.generate(
            pageCount, (index) => Pointer.fromAddress(0),
            growable: false),
        disposeCallback: disposeCallback,
      );
    });
  }

  @override
  Future<PdfPage> getPage(int pageNumber) async {
    var page = _pages[pageNumber - 1];
    if (page.address == 0) {
      page = _pages[pageNumber - 1] = pdfium.FPDF_LoadPage(doc, pageNumber);
      if (page.address == 0) throw Exception('Failed to load page $pageNumber');
    }
    return PdfPagePdfium._(
      document: this,
      pageNumber: pageNumber,
      width: pdfium.FPDF_GetPageWidthF(page),
      height: pdfium.FPDF_GetPageHeightF(page),
    );
  }

  @override
  Future<void> dispose() async {
    pdfium.FPDF_CloseDocument(doc);
    disposeCallback?.call();
  }

  @override
  bool operator ==(covariant PdfDocumentPdfium other) {
    if (identical(this, other)) return true;

    return other.doc == doc;
  }

  @override
  int get hashCode => doc.hashCode;
}

class PdfPagePdfium extends PdfPage {
  PdfPagePdfium._({
    required super.document,
    required super.pageNumber,
    required super.width,
    required super.height,
  });

  @override
  PdfDocumentPdfium get document => super.document as PdfDocumentPdfium;

  FpdfPdfPage get _page => document._pages[pageNumber - 1];

  @override
  Future<PdfPageImage> render(
      {int x = 0,
      int y = 0,
      int? width,
      int? height,
      double? fullWidth,
      double? fullHeight,
      bool backgroundFill = true,
      bool allowAntialiasingIOS = false}) async {
    fullWidth ??= this.width;
    fullHeight ??= this.height;
    width ??= fullWidth.toInt();
    height ??= fullHeight.toInt();

    final buffer = malloc.allocate<Uint8>(width * height * 4);
    final c = backgroundFill ? 0xff : 0;
    final size = width * height * 4;
    for (int i = 0; i < size; i++) {
      buffer[i] = c;
    }
    final bmp = pdfium.FPDFBitmap_CreateEx(
        width, height, FPDFBitmap_BGRA, buffer.cast<Void>(), width * 4);
    pdfium.FPDF_RenderPageBitmap(bmp, _page, x, y, width, height, 0, 0);
    pdfium.FPDFBitmap_Destroy(bmp);

    return PdfPageImagePdfium._(
        pageNumber: pageNumber,
        x: x,
        y: y,
        width: width,
        height: height,
        fullWidth: fullHeight ?? this.width,
        fullHeight: fullHeight ?? this.height,
        pageWidth: this.width.toDouble(),
        pageHeight: this.height.toDouble(),
        buffer: buffer);
  }

  @override
  Future<bool> renderOnTexture(
      {required PdfPageImageTexture texture,
      int x = 0,
      int y = 0,
      required int width,
      required int height,
      double? fullWidth,
      double? fullHeight,
      bool backgroundFill = true,
      bool allowAntialiasingIOS = true}) async {
    final tex = texture as PdfPageImageTextureMethodChannel;
    fullWidth ??= this.width;
    fullHeight ??= this.height;
    tex.paint(
      width,
      height,
      backgroundFill: backgroundFill,
      paint: (pixels, width, height, stride, size) {
        final bmp = pdfium.FPDFBitmap_CreateEx(
            width, height, FPDFBitmap_BGRA, pixels.cast<Void>(), stride);
        pdfium.FPDF_RenderPageBitmap(
          bmp,
          _page,
          x,
          y,
          fullWidth!.toInt(),
          fullHeight!.toInt(),
          0,
          FPDF_REVERSE_BYTE_ORDER,
        );
        pdfium.FPDFBitmap_Destroy(bmp);
      },
    );
    return true;
  }
}

class PdfPageImagePdfium extends PdfPageImage {
  PdfPageImagePdfium._({
    required super.pageNumber,
    required super.x,
    required super.y,
    required super.width,
    required super.height,
    required super.fullWidth,
    required super.fullHeight,
    required super.pageWidth,
    required super.pageHeight,
    required Pointer<Uint8> buffer,
  }) : _buffer = buffer;

  Pointer<Uint8>? _buffer;
  ui.Image? _imageCached;

  @override
  Pointer<Uint8>? get buffer => _buffer;

  @override
  Future<ui.Image> createImageDetached() => _decodeRgba(width, height, pixels);

  @override
  Future<ui.Image> createImageIfNotAvailable() async {
    _imageCached ??= await _decodeRgba(width, height, pixels);
    return _imageCached!;
  }

  @override
  void dispose() {
    _imageCached?.dispose();
    _imageCached = null;
    if (_buffer != null) {
      calloc.free(_buffer!);
      _buffer = null;
    }
  }

  @override
  ui.Image? get imageIfAvailable => _imageCached;

  @override
  Uint8List get pixels => _buffer!.asTypedList(width * height * 4);

  /// Decode BGRA raw image from native code.
  static Future<ui.Image> _decodeRgba(int width, int height, Uint8List pixels) {
    final comp = Completer<ui.Image>();
    ui.decodeImageFromPixels(pixels, width, height, ui.PixelFormat.bgra8888,
        (image) => comp.complete(image));
    return comp.future;
  }
}
